/***************************************************************************
 *   $Id: variable.c,v 1.1.1.1 2011-11-17 16:42:29 scg Exp $
 *                                                                         *
 *   (C)2011 St(u)dio of Computer Games                                    *
 *   Alexander Ozumenko <a.ozumenko@gmail.com>                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

#include "variable.h"
#include "list.h"
#include "hash.h"



/* expand enviroments variables */
static inline int enviroment_expand(struct enviroment *env, enum VARIABLE_TYPE type) {
    struct variable *var;
    int i;

    for(i = 0; i < VARIABLE_PAGE; i++) {
        var = malloc(sizeof(struct variable));
        if(var == NULL) {
            env->error = errno;
            return -1;
        }

        var->env = env;
        var->type = type;

        switch(type) {
        case VARIABLE_TYPE_SCALAR:
            var->scalar.value = NULL;
            break;
        case VARIABLE_TYPE_ARRAY:
            INIT_LIST_HEAD(&var->array.list);
            break;
        case VARIABLE_TYPE_HASH:
#if 0
            var->hash.hash.table = NULL;
#else
            hash_init(&var->hash.hash, 256);   // FixMe: 256 to #define
#endif
            break;
        default:
            break;
        }

        list_add_tail(&var->placement.list, &env->var_free[type]);
    }

    return 0;
}


/* get next free variable */
static inline struct variable *variable_get_free(struct enviroment *env, enum VARIABLE_TYPE type) {
    struct variable *var;
    struct list_head *var_free = &env->var_free[type];

    if(list_empty(var_free)) {
        if(enviroment_expand(env, type) < 0) {
            return NULL;
	}
    }

    var = list_entry(var_free->next, struct variable, placement.list);
    list_del(&var->placement.list);

#if 0
    switch(type) {
    case VARIABLE_TYPE_HASH:
        if(var->hash.hash.table == NULL) {
            hash_init(&var->hash.hash, 256);   // FixMe: 256 to #define
        }
        break;
    default:
        break;
    }
#endif

    return var;
}


static inline void array_push(struct variable *var, va_list ap) {
    struct variable *value;

    while((value = va_arg(ap, struct variable *)) != NULL) {
        list_add_tail(&value->placement.list, &var->array.list);
    }
}


static inline int hash_set(struct variable *var, va_list ap) {
    const char *key;
    struct variable *value;
    struct hash_entry *hash_entry;

    while((key = va_arg(ap, char *)) != NULL) {
        hash_entry = hash_del(&var->hash.hash, key);
        if(hash_entry != NULL ) {
            variable_release(hash_get_entry(hash_entry, struct variable, placement.hash));
        }

        value = va_arg(ap, struct variable *);
        if(value != NULL) {
            if(hash_add(&var->hash.hash, key, &value->placement.hash)) {
                var->env->error = errno;
                return -1;
            }
        }
    }

    return 0;
}



/*
 * ENVIROMENT interface section
 */

/* enviroment init */
int enviroment_init(struct enviroment *env) {
    int i;

    env->error = 0;

    for(i = 0; i < VARIABLE_TYPE_MAX; i++) {
        INIT_LIST_HEAD(&env->var_free[i]);
    }

    env->root = variable_get_free(env, VARIABLE_TYPE_HASH);
    if(env->root == NULL) {
        return -1;
    }

    return 0;
}


/* enviroment release */
void enviroment_release(struct enviroment *env) {
    struct variable *var;
    int i;

    /* release root HASH */
    variable_release(env->root);

    /* free all HASHes */
    list_for_each_entry(var, &env->var_free[VARIABLE_TYPE_HASH], placement.list) {
        /* after variable_release() all hashes must be empty,
           NULL at last parameter should not lead to a crash */
        hash_release(&var->hash.hash, NULL);
    }

    /* free variables */
    for(i = 0; i < VARIABLE_TYPE_MAX; i++) {
        struct list_head *var_free = &env->var_free[i];

        while(!list_empty(var_free)) {
            var = list_entry(var_free->next, struct variable, placement.list);
            list_del(var_free->next);
            free(var);
        }
    }
}


/* get enviroment's root variable */
struct variable *enviroment_get_root(struct enviroment *env) {
    return env->root;
}


/* get last error */
int enviroment_errno(struct enviroment *env) {
    return env->error;
}



/*
 * Variables common section
 */

/* release variable */
void variable_release(struct variable *var) {
    /* variable is not attachet to LIST or HASH */
    if(var->placement.list.prev != NULL || var->placement.list.next != NULL) {
        return;
    }

    switch(var->type) {
    case VARIABLE_TYPE_SCALAR:
        free(var->scalar.value);
        break;

    case VARIABLE_TYPE_ARRAY:
        {
            while(!list_empty(&var->array.list)) {
                struct list_head *pos;

                pos = var->array.list.next;
                list_del(pos);

                variable_release(list_entry(pos, struct variable, placement.list));
            }
        }
        break;

    case VARIABLE_TYPE_HASH:
        {
            struct list_head released;

            INIT_LIST_HEAD(&released);
            hash_clean(&var->hash.hash, &released);

            while(!list_empty(&released)) {
                struct hash_entry *hash_entry;
                struct list_head *pos;

                pos = released.next;
                list_del(pos);

                hash_entry = list_entry(pos, struct hash_entry, list);
                variable_release(hash_get_entry(hash_entry, struct variable, placement.hash));
            }
        }
        break;

    default:
        break;
    }

    list_add_tail(&var->placement.list, &var->env->var_free[var->type]);
}


/* get variable type */
inline enum VARIABLE_TYPE variable_type(struct variable *var) {
    return var->type;
}



/*
 * SCALAR interface section
 */

/* create SCALAR variable */
struct variable *variable_scalar(struct enviroment *env, const char *value) {
    struct variable *var;

    if(value == NULL) {
        return NULL;
    }

    var = variable_get_free(env, VARIABLE_TYPE_SCALAR);
    if(var == NULL) {
        return NULL;
    }

    var->scalar.value = strdup(value);
    if(var->scalar.value == NULL) {
        env->error = errno;
        return NULL;
    }

    return var;
}


/* get SCALAR variable value */
inline const char *variable_scalar_get(struct variable *var) {
    if(var == NULL) {
        return NULL;
    }

    if(var->type != VARIABLE_TYPE_SCALAR) {
        var->env->error = EINVAL;
        return NULL;
    }

    return var->scalar.value;
}



/*
 * ARRAY interface section
 */

/* create ARRAY variable */
struct variable *variable_array(struct enviroment *env, ...) {
    struct variable *var;
    va_list ap;

    var = variable_get_free(env, VARIABLE_TYPE_ARRAY);
    if(var == NULL) { 
        return NULL;
    }

    va_start(ap, env);
    array_push(var, ap);
    va_end(ap);

    return var;
}


/* push variables into the array */
struct variable *variable_array_push(struct variable *var, ...) {
    va_list ap;

    if(var == NULL) {
        return NULL;
    }

    if(var->type != VARIABLE_TYPE_ARRAY) {
        var->env->error = EINVAL;
        return NULL;
    }

    va_start(ap, var);
    array_push(var, ap);
    va_end(ap);

    return var;
}



/*
 * HASH interface section
 */

/* create HASH variable */
struct variable *variable_hash(struct enviroment *env, ...) {
    struct variable *var;
    va_list ap;

    var = variable_get_free(env, VARIABLE_TYPE_HASH);
    if(var == NULL) {
        return NULL;
    }

    va_start(ap, env);
    hash_set(var, ap);
    va_end(ap);

    return var;
}


/* add variable to hash */
struct variable *variable_hash_set(struct variable *var, ...) {
    va_list ap;

    if(var == NULL) {
        return NULL;
    }

    if(var->type != VARIABLE_TYPE_HASH) {
        var->env->error = EINVAL;
        return NULL;
    }

    va_start(ap, var);
    hash_set(var, ap);
    va_end(ap);

    return var;
}


/* get varable from hash */
struct variable *variable_hash_get(struct variable *var, const char *key) {
    struct hash_entry *hash_entry;

    if(var == NULL) {
        return NULL;
    }

    if(var->type != VARIABLE_TYPE_HASH) {
        var->env->error = EINVAL;
        return NULL;
    }

    hash_entry = hash_get(&var->hash.hash, key);
    if(hash_entry == NULL) {
        return NULL;
    }

    return hash_get_entry(hash_entry, struct variable, placement.hash);
}
